<?php
/**
* @package msta calendar Extension
*/

namespace msta\calendar\includes;

class CalendarImporter extends CalendarBase {
  protected $auth;
  private $phpbb_root_path;

  public function __construct (\phpbb\db\driver\driver_interface $db, \phpbb\config\config $config, \phpbb\user $user, \phpbb\auth\auth $auth, $phpbb_root_path) {
    parent::__construct($db, $config, $user);
    $this->auth = $auth;
    $this->phpbb_root_path = $phpbb_root_path;
  }

  /*
   readBlock
   Recursive parser for parseICS.
  */
  private function readBlock($str, &$i) {
    $o = array();
    $l = ''; // last variable name in o
    while ($i < count($str)) {
      if (substr($str[$i], 0, 1) == ' ' && $l != '') {
        $o[$l] .= str_replace(array_reverse($this->icsSpecialCharsEscaped), array_reverse($this->icsSpecialChars), substr($str[$i], 1));
        $i++;
        continue;
      }
      $s = explode(':', $str[$i]);
      $i++;
      if ($s[0] == 'END') return $o;
      else if ($s[0] == 'BEGIN') {
        if (!$o[$s[1]]) $o[$s[1]] = array();
        $o[$s[1]][] = $this->readBlock($str, $i);
      } else {
        $l = explode(';', $s[0]);
        for ($j = 1; $j < count($l); $j++) {
          $l[$j] = explode('=', $l[$j]);
          if (count($l[$j]) == 2) $o[$l[0].'_'.$l[$j][0]] = $l[$j][1];
        }
        $l = $l[0];
        array_shift($s);
        $o[$l] = str_replace(array_reverse($this->icsSpecialCharsEscaped), array_reverse($this->icsSpecialChars), join(':', $s));
      }
    }
    return $o;
  }

  /*
   parseICS
   Transforms an ICS-file given as a string into an array-structure holding the information.
  */
  public function parseICS ($str) {
    $str = explode("\n", join('', explode("\r", $str)));
    $i = 0;
    return $this->readBlock($str, $i);
  }

  private function applyFilter ($categories, $promoter, $field, $map) {
    $categories = explode(',', $categories);
    $map = explode('|', $map);
    $dst = 0;
    for ($i = 0; $i < count($map); $i++) {
      if ($map[$i] == '') continue;
      if ($map[$i] == 'auto') {
        if ($field == 'category')
          for ($j = 0; $j < count($categories); $j++) {
            $dst = $this->categoryIdByName($categories[$j]);
            if ($dst != 0) break 2;
          }
        else {
          $dst = $this->promoterIdByName($promoter);
          if ($dst != 0) break;
        }
      } else {
        $spl = explode('=', $map[$i]);
        if ($spl[0] == '*' && count($spl) == 2) {
          $dst = intval($spl[1]);
          break;
        }
        if (count($spl) != 3) continue;
        $spl[1] = urldecode($spl[1]);
        if ($spl[0] == 'cat') {
          for ($j = 0; $j < count($categories); $j++)
            if ($spl[1] == $categories[$j]) {
              $dst = intval($spl[2]);
              break 2;
            }
        } else if ($spl[0] == 'pro') {
          if ($spl[1] == $promoter) {
            $dst = intval($spl[2]);
            break;
          }
        }
      }
    }
    if ($dst != 0) {
      if ($field == 'category' && $this->categoryInfo($dst)['name'] == '') $dst = 0;
      if ($field == 'promoter' && $this->promoterInfo($dst)['group_name'] == '') $dst = 0;
    }
    return $dst;
  }

  private function getIcsDate ($icsEvent, $dateKey, &$date, &$time) {
    $dateType = 'DATE-TIME';
    if (array_key_exists($dateKey.'_VALUE', $icsEvent))
      $dateType = $icsEvent[$dateKey.'_VALUE'];
    if ($dateType == 'DATE') {
      if (strlen($icsEvent[$dateKey]) != 8) return false;
      $time = '';
      $date = substr($icsEvent[$dateKey], 0, 4).'-'.substr($icsEvent[$dateKey], 4, 2).'-'.substr($icsEvent[$dateKey], 6, 2);
      if ($dateKey == 'DTEND') {
        $dt = new \DateTime($date.' 12:00', $this->utc_timezone);
        $dt->sub(new \DateInterval('P1D'));
        $date = $dt->format('Y-m-d');
      }
    } else if ($dateType == 'DATE-TIME') {
      if (strlen($icsEvent[$dateKey]) < 13 || substr($icsEvent[$dateKey], 8, 1) != 'T') return false;
      $time = substr($icsEvent[$dateKey], 9, 2).':'.substr($icsEvent[$dateKey], 11, 2);
      $date = substr($icsEvent[$dateKey], 0, 4).'-'.substr($icsEvent[$dateKey], 4, 2).'-'.substr($icsEvent[$dateKey], 6, 2);
      if (array_key_exists($dateKey.'_TZID', $icsEvent)) {
        try {
          $tzid = explode('/', $icsEvent[$dateKey.'_TZID']);
          while (count($tzid) > 2) array_shift($tzid); // sometimes, there are unneccessary information in tzid...
          $tzid = join('/', $tzid);
          if ($tzid == 'W. Europe Standard Time') $tzid = 'Europe/Berlin'; // special case...
          $tz = new \DateTimeZone($tzid);
        } catch (Exception $e) {
          return false;
        }
        $dt = new \DateTime($date.' '.$time, $tz);
        $dt->setTimezone($this->utc_timezone);
        $date = $dt->format('Y-m-d');
        $time = $dt->format('H:i');
      }
    }
    return true;
  }

  /*
   insertOrUpdateEvent
   Adds the date into the table, removes deleted-entry for this post_id, if exists.
   date: The new date as a validated (see checkEvent) event object.
   isNew: It must be tested before, weather a date for date['post_id'] already exists. If not, this parameter must be true.
  */
  public function insertOrUpdateEvent ($date, $isNew, $modTime=0) {
    $utfDate = array();
    foreach ($date as $k => $v)
      if (gettype($v) == 'string') $utfDate[$k] = utf8_encode_ucr($v);
      else $utfDate[$k] = $v;
    if ($isNew) { // new date
      if ($this->db->sql_query('INSERT INTO '.$this->table_prefix.'calendar '.$this->db->sql_build_array('INSERT', $utfDate)))
        $notify = 'calendar_notify_add';
    } else { // update existing date
      if ($this->db->sql_query('UPDATE '.$this->table_prefix.'calendar SET '.$this->db->sql_build_array('UPDATE', $utfDate).' WHERE post_id = '.$date['post_id']))
        $notify = 'calendar_notify_change';
    }
    $this->db->sql_query('DELETE FROM '.$this->table_prefix."calendar_deleted WHERE post_id = ".intval($date['post_id']));
    if ($modTime == 0) $modTime = date('U');
    $this->db->sql_query('UPDATE '.$this->table_prefix."posts SET post_edit_time = ".$modTime." WHERE post_id = ".intval($date['post_id']));
  }

  /*
   deleteEvent
   Deletes the event with given ID and creates a deleted-entry for it.
   Returns true, if the event existed before (Can be called for dates, which don't exist and has no effect then).
  */
  public function deleteEvent ($post_id, $manually=true, $deleteLinkedPost=false) {
    $existingRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT cal.post_id, cal.uid, p.topic_id FROM '.$this->table_prefix.'calendar AS cal LEFT JOIN '.$this->table_prefix.'posts AS p ON p.post_id = cal.post_id WHERE cal.post_id = '.intval($post_id)));
    if ($existingRow) {
      $nowDate = new \DateTime();
      $this->db->sql_query('DELETE FROM '.$this->table_prefix.'calendar WHERE post_id = '.intval($existingRow['post_id']));
      $this->db->sql_query('INSERT INTO '.$this->table_prefix.'calendar_deleted (post_id, uid, delete_time, deleted_by_user) VALUES ('.intval($existingRow['post_id']).", '".$this->db->sql_escape($existingRow['uid'])."', ".$nowDate->getTimestamp().", ".(($manually)?1:0).");");
      if ($deleteLinkedPost) {
        $this->db->sql_query('DELETE FROM '.$this->table_prefix.'posts WHERE post_id = '.intval($existingRow['post_id']));
        if (intval($existingRow['topic_id']) > 0 && $this->db->sql_affectedrows() > 0)
          $this->db->sql_query('UPDATE '.$this->table_prefix.'topics SET topic_posts_approved = topic_posts_approved - 1 WHERE topic_id = '.intval($existingRow['topic_id']));
      }
      return true;
    } else return false;
  }

  private function parseRrule(&$evVar, $rrule) {
    if ($rrule == '') return;
    $hasUntil = false;
    $rrule = explode(';', $rrule);
    for ($k = 0; $k < count($rrule); $k++) {
      $rrule[$k] = explode('=', strtoupper($rrule[$k]));
      if (count($rrule[$k]) == 1) continue;
      switch ($rrule[$k][0]) {
        case 'FREQ':
          if ($rrule[$k][1] == 'MONTHLY' && $evVar['repeat_mode'] == 0) $evVar['repeat_mode'] = 1;
          if ($rrule[$k][1] == 'MONTHLY' && $evVar['repeat_mode'] == -3) $evVar['repeat_mode'] = 3;
          if ($rrule[$k][1] == 'DAILY') $evVar['repeat_mode'] = 2;
          break;
        case 'UNTIL':
          if (strlen($rrule[$k][1]) != 8) break;
          $y = intval(substr($rrule[$k][1], 0, strlen($rrule[$k][1])-4));
          $m = intval(substr($rrule[$k][1], strlen($rrule[$k][1])-4, 2));
          $d = intval(substr($rrule[$k][1], strlen($rrule[$k][1])-2, 2));
          $evVar['date_to'] = str_pad($y, 4, '0', STR_PAD_LEFT).'-'.str_pad($m, 2, '0', STR_PAD_LEFT).'-'.str_pad($d, 2, '0', STR_PAD_LEFT);
          $hasUntil = true;
          break;
        case 'INTERVAL':
          if ($evVar['repeat_nr'] == 0) $evVar['repeat_nr'] = intval($rrule[$k][1]);
          else $evVar['repeat_nr'] = -intval($rrule[$k][1]);
          break;
        case 'BYDAY':
          if ($evVar['repeat_mode'] == 0) $evVar['repeat_mode'] = -3; // will be set to +3 when FREQ is monthly.
          else if ($evVar['repeat_mode'] == 1) $evVar['repeat_mode'] = 3;
          if (intval($rrule[$k][1]) < 0) {
            if ($evVar['repeat_nr'] == 0) $evVar['repeat_nr'] = PHP_INT_MIN; // just remember to make value negative for INTERVAL parsing
            else $evVar['repeat_nr'] = -$evVar['repeat_nr'];
          }
          break;
      }
    }
    if ($evVar['repeat_mode'] < 0) $evVar['repeat_mode'] = 0; // invalid combination of rules
    if (!$hasUntil && $evVar['repeat_mode'] > 0) $evVar['date_to'] = '';
  }

  /*
   importAsOwnEvents
   Imports events from an ICS-Structure into the forum as if they were created in the forum.
   ics: the parsed ICS structure.
   defaultTopicId: The id of an existing topic, where the dates' posts will be created. If negative sign, this topic will be enforces as absolute value, otherwise dates, that has a link to an existing post, will be bound to this post.
   promoterId: Id of the promoter group.
   defaultCategory: if of the default category. If negative sign, this category will be enforced as absolute value.
   public - flags:
    1 = public default is yes
    2 = override ics with default
  */
  public function importEvents ($ics, $settings) { // $defaultTopicId, $promoterId, $defaultCategory, $public
    if (!$ics['VCALENDAR'] || !$ics['VCALENDAR'][0]['VEVENT']) return array('ICS contains no events!');
    $dateFrom = '';
    $timeFrom = '';
    $dateTo = '';
    $timeTo = '';
    $report = array();
    $numImported = 0;
    $numUpdated = 0;
    $numUpToDate = 0;
    $nowTimestamp = date('U');
    $checksum = md5('[...]');
    $regardedIDs = array();
    // TODO call some function, that comverts events without post into deleted entries
    // get default topic's information
    $topicRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT topic_id, forum_id FROM '.$this->table_prefix.'topics WHERE topic_id = '.intval($settings['topic_id'])));
    if (!$topicRow) return array('Cannot find destination topic!');
    // add events
    $events = $ics['VCALENDAR'][0]['VEVENT'];
    $report[] = count($events).' events found in ICS.';
    for ($i = 0; $i < count($events); $i++) {
      if (!$this->getIcsDate($events[$i], 'DTSTART', $dateFrom, $timeFrom) || !$this->getIcsDate($events[$i], 'DTEND', $dateTo, $timeTo)) {
        $report[] = 'Invalid DTSTART or DTEND value at '.htmlspecialchars($events[$i]['SUMMARY']);
        continue;
      }
      $evVar = array(
        'post_id' => 0,
        'event_name' => htmlspecialchars($events[$i]['SUMMARY']),
        'location' => htmlspecialchars($events[$i]['LOCATION']),
        'url' => $events[$i]['URL'],
        'description' => htmlspecialchars($events[$i]['DESCRIPTION']),
        'category' => $this->applyFilter(htmlspecialchars($events[$i]['CATEGORIES']), htmlspecialchars($events[$i]['ORGANIZER_CN']), 'category', $settings['category_mapping']),
        'promoter' => $this->applyFilter(htmlspecialchars($events[$i]['CATEGORIES']), htmlspecialchars($events[$i]['ORGANIZER_CN']), 'promoter', $settings['promoter_mapping']),
        'public' => ((($settings['publicity'] & 2) == 0 && $events[$i]['CLASS'] && $events[$i]['CLASS'] == 'PUBLIC') || ((($settings['publicity'] & 2) == 2 || !$events[$i]['CLASS']) && ($settings['publicity'] & 1) == 1))?1:0,
        'date_from' => $dateFrom,
        'time_from' => $timeFrom,
        'date_to' => $dateTo,
        'time_to' => $timeTo,
        'repeat_mode' => 0,
        'repeat_nr' => 0,
        'canceled' => (strtoupper($events[$i]['STATUS']) == 'CANCELLED')?1:0,
        'anniversary' => ((($settings['anniversarity'] & 2) == 0 && $timeFrom == '' && $timeTo == '' && $dateTo < $dateFrom) || (($settings['anniversarity'] & 2) == 2 && ($settings['anniversarity'] & 1) == 1))?1:0,
        'uid' => ($settings['keep_uid'] == 1)?$events[$i]['UID']:'',
      );
      if ($evVar['category'] == 0) continue; // no category filter did match; don't import this event
      if (array_key_exists('RRULE', $events[$i])) $this->parseRrule($evVar, $events[$i]['RRULE']);
      $this->checkEvent($evVar);
      if ($evVar['date_from'] == '') {
        $report[] = 'Invalid beginning date at '.htmlspecialchars($events[$i]['SUMMARY']);
        continue;
      }
      $existingDate = null;
      if ($evVar['uid'] != '') { // check, if the date belongs to an existing date
        $existingDate = $this->db->sql_fetchrow($this->db->sql_query('SELECT cal.*, p.post_edit_time, p.post_time FROM '.$this->table_prefix.'calendar AS cal INNER JOIN '.$this->table_prefix."posts AS p ON p.post_id=cal.post_id WHERE cal.uid = '".$this->db->sql_escape($evVar['uid'])."'"));
        if ($existingDate) $evVar['post_id'] = $existingDate['post_id'];
        else {
          $deletedDate = $this->db->sql_fetchrow($this->db->sql_query('SELECT * FROM '.$this->table_prefix."calendar_deleted WHERE uid = '".$this->db->sql_escape($evVar['uid'])."'"));
          if ($deletedDate && $deletedDate['deleted_by_user'] == 1) continue; // the date has been deleted by a forum user. Don't re-import it!
        }
      }
      if ($evVar['post_id'] == 0 && $settings['search_fitting_post'] == 1 && strpos($evVar['url'], generate_board_url().'/viewtopic.php?') === 0) { // check, if the date belongs to an existing post
        $query = explode('#', substr($evVar['url'], strlen($this->phpbb_root_path)+14));
        if (count($query) == 2 && substr($query[1], 0, 1) == 'p') $evVar['post_id'] = intval(substr($query[1], 1));
        else {
          $query = explode('&', $query[0]);
          $topicId = 0;
          for ($j = 0; $j < count($query); $j++) {
            $sp = explode('=', $query[$j]);
            if (count($sp) != 2) continue;
            if ($sp[0] == 'p') {
              $evVar['post_id'] = intval($sp[1]);
              break;
            }
            if ($sp[0] == 't')
              $topicId = intval($sp[1]);
          }
          if ($evVar['post_id'] == 0 && $topicId != 0) { // get first post of topic
            $postTopicRow = $this->db->sql_fetchrow($this->db->sql_query('SELECT topic_first_post_id FROM '.$this->table_prefix.'topics WHERE topic_id = '.$topicId));
            if ($postTopicRow)
              $evVar['post_id'] = intval($postTopicRow['topic_first_post_id']);
          }
        }
      }
      if ($evVar['post_id'] == 0) { // don't belong to an existing post; create a new one in default topic for guest account
        $lastId = 0;
        if ($this->db->sql_query('INSERT INTO '.$this->table_prefix.'posts (topic_id, forum_id, poster_id, post_time, post_visibility, post_subject, post_text, post_checksum) VALUES ('.intval($topicRow['topic_id']).', '.intval($topicRow['forum_id']).', 1, '.$nowTimestamp.", 1, '".$this->db->sql_escape(utf8_encode_ucr($evVar['event_name']))."', '[...]', '".$checksum."')"))
          $lastId = intval($this->db->sql_nextid());
        if ($lastId == 0) {
          $report[] = 'Cannot create new post for '.htmlspecialchars($events[$i]['SUMMARY']);
          continue;
        }
        $this->db->sql_query('UPDATE '.$this->table_prefix.'topics SET topic_last_post_id = '.$lastId.', topic_last_poster_id = '.intval($this->user->data['user_id']).", topic_last_poster_name = 'cal', topic_last_poster_colour = '', topic_last_post_subject = '".$this->db->sql_escape(utf8_encode_ucr($evVar['event_name']))."', topic_last_post_time = ".$nowTimestamp.', topic_posts_approved = topic_posts_approved + 1 WHERE topic_id = '.intval($topicRow['topic_id']));
        $evVar['post_id'] = $lastId;
      }
      if (!$existingDate) $existingDate = $this->db->sql_fetchrow($this->db->sql_query('SELECT cal.*, p.post_edit_time, p.post_time FROM '.$this->table_prefix.'calendar AS cal INNER JOIN '.$this->table_prefix.'posts AS p ON p.post_id=cal.post_id WHERE cal.post_id = '.$evVar['post_id']));
      $regardedIDs[] = $evVar['post_id'];
      if ($existingDate) {
        $lastUpdate = 0;
        if ($existingDate['post_time']) $lastUpdate = intval($existingDate['post_time']);
        if ($existingDate['post_edit_time']) $lastUpdate = intval($existingDate['post_edit_time']);
        if ($lastUpdate > 0) {
          $modDate = '';
          $modTime = '';
          if ($this->getIcsDate($events[$i], 'DTSTART', $modDate, $modTime) && $modDate != '') {
            $modDt = new \DateTime($modDate.' '.$modTime, $this->utc_timezone);
            if ($modDt->getTimestamp() < $lastUpdate) {
              $numUpToDate++;
              continue; // the latest date modification (by user or auto-import) is newer than the last mod of the ics event.
            }
          }
        }
      }
      $this->insertOrUpdateEvent($evVar, !$existingDate, $nowTimestamp);
      if ($existingDate) $numUpdated++;
      else $numImported++;
    }
    $report[] = $numImported.' events imported.';
    if ($numUpdated > 0) $report[] = $numUpdated.' events updated.';
    if ($numUpToDate > 0) $report[] = $numUpToDate.' events have a newer modification date in the forum and remain unchainged.';
    // delete all events from destination topic, that dont belong to to the import
    $numDeleted = 0;
    $nowDate = date('Y-m-d');
    $eventsToDelete = $this->db->sql_query('SELECT cal.post_id, cal.date_from, p.topic_id, p.post_text FROM '.$this->table_prefix.'calendar AS cal INNER JOIN '.$this->table_prefix.'posts AS p ON p.post_id=cal.post_id WHERE p.topic_id = '.intval($topicRow['topic_id']).((count($regardedIDs) > 0)?' AND p.post_id NOT IN ('.join(', ', $regardedIDs).')':''));
    while ($row = $this->db->sql_fetchrow($eventsToDelete)) {
      if ($settings['keep_past_events'] == 1 && $row['date_from'] <= $nowDate) continue;
      if ($this->deleteEvent($row['post_id'], false, $row['post_text'] == '[...]'))
        $numDeleted++;
    }
    if ($numDeleted > 0) $report[] = $numDeleted.' events deleted.';
    $this->db->sql_query('UPDATE '.$this->table_prefix.'calendar_external SET last_fetch = '.$nowTimestamp.' WHERE topic_id = '.intval($settings['topic_id']));
    $report[] = 'last fetch = '.$nowTimestamp;
    return $report;
  }
} 
